@@ -144,6 +144,19 @@ init_cygheap::init_installation_root ()
     api_fatal ("Can't initialize Cygwin installation root dir.\n"
 	       "GetModuleFileNameW(%p, %p, %u), %E",
 	       cygwin_hmodule, installation_root_buf, PATH_MAX);
+
+  /* We don't care if fetching the final pathname fails, it's non-fatal and
+     the path returned by GetModuleFileNameW is still valid. */
+  HANDLE h;
+  h = CreateFileW (installation_root_buf, GENERIC_READ, FILE_SHARE_VALID_FLAGS,
+		   &sec_none, OPEN_EXISTING, 0, 0);
+  if (h != INVALID_HANDLE_VALUE)
+    {
+      GetFinalPathNameByHandleW (h, installation_root_buf, PATH_MAX,
+				 FILE_NAME_NORMALIZED);
+      CloseHandle (h);
+    }
+
   PWCHAR p = installation_root_buf;
   if (wcsncasecmp (p, L"\\\\", 2))	/* Normal drive letter path */
     {
@@ -170,6 +183,11 @@ init_cygheap::init_installation_root ()
 	  if (p)
 	    p = wcschr (p + 1, L'\\');  /* Skip share name */
 	}
+      else /* Long path prefix followed by drive letter path */
+	{
+	  len = 4;
+	  p += 4;
+	}
     }
   installation_root_buf[1] = L'?';
   RtlInitEmptyUnicodeString (&installation_key, installation_key_buf,
@@ -626,7 +644,7 @@ init_cygheap::add_tls (_cygtls *t)
      still utilizes the thread's _cygtls area, things go awry.
 
      The following methods take this into account:
-     
+
      - The thread mutex is generally only locked under tls_sentry locking.
      - remove_tls, called from _cygtls::remove, locks the mutex before
        removing the threadlist entry and _cygtls::remove then unlocks and
