@@ -79,7 +79,7 @@ extern "C" void yield ();
 
 #define import_address(x) __import_address ((void *)(x))
 void * __reg1 __import_address (void *);
- 
+
 #define caller_return_address() \
 		__caller_return_address (__builtin_return_address (0))
 void * __reg1 __caller_return_address (void *);
@@ -98,6 +98,9 @@ transform_chars (PUNICODE_STRING upath, USHORT start_idx)
 
 PWCHAR transform_chars_af_unix (PWCHAR, const char *, __socklen_t);
 
+/* Get handle count of an object. */
+ULONG get_obj_handle_count (HANDLE h);
+
 /* Memory checking */
 int __reg2 check_invalid_virtual_addr (const void *s, unsigned sz);
 
