@@ -55,10 +55,11 @@ void
 pinfo::thisproc (HANDLE h)
 {
   procinfo = NULL;
+  bool execed = !!h;
 
   DWORD flags = PID_IN_USE | PID_ACTIVE;
   /* Forked process or process started from non-Cygwin parent needs a pid. */
-  if (!h)
+  if (!execed)
     {
       cygheap->pid = create_cygwin_pid ();
       flags |= PID_NEW;
@@ -72,7 +73,8 @@ pinfo::thisproc (HANDLE h)
   procinfo->dwProcessId = myself_initial.dwProcessId;
   procinfo->sendsig = myself_initial.sendsig;
   wcscpy (procinfo->progname, myself_initial.progname);
-  create_winpid_symlink ();
+  if (!execed)
+    create_winpid_symlink ();
   procinfo->exec_sendsig = NULL;
   procinfo->exec_dwProcessId = 0;
   debug_printf ("myself dwProcessId %u", procinfo->dwProcessId);
@@ -106,6 +108,8 @@ pinfo_init (char **envp, int envc)
 
   myself->process_state |= PID_ACTIVE;
   myself->process_state &= ~(PID_INITIALIZING | PID_EXITED | PID_REAPED);
+  if (being_debugged ())
+    myself->process_state |= PID_DEBUGGED;
   myself.preserve ();
   debug_printf ("pid %d, pgid %d, process_state %y",
 		myself->pid, myself->pgid, myself->process_state);
@@ -156,6 +160,9 @@ pinfo::status_exit (DWORD x)
 	 a lengthy small_printf instead. */
       x = SIGBUS;
       break;
+    case STATUS_CONTROL_C_EXIT:
+      x = SIGINT;
+      break;
     default:
       debug_printf ("*** STATUS_%y\n", x);
       x = 127 << 8;
@@ -1526,23 +1533,35 @@ winpids::enum_processes (bool winpid)
 
   if (!winpid)
     {
+      tmp_pathbuf tp;
+      NTSTATUS status;
       HANDLE dir = get_shared_parent_dir ();
       BOOLEAN restart = TRUE;
-      ULONG context;
-      struct fdbi
+      bool last_run = false;
+      ULONG context = 0;
+      PDIRECTORY_BASIC_INFORMATION dbi_buf = (PDIRECTORY_BASIC_INFORMATION)
+					     tp.w_get ();
+      while (!last_run)
 	{
-	  DIRECTORY_BASIC_INFORMATION dbi;
-	  WCHAR buf[2][NAME_MAX + 1];
-	} f;
-      while (NT_SUCCESS (NtQueryDirectoryObject (dir, &f, sizeof f, TRUE,
-						 restart, &context, NULL)))
-	{
-	  restart = FALSE;
-	  f.dbi.ObjectName.Buffer[f.dbi.ObjectName.Length / sizeof (WCHAR)] = L'\0';
-	  if (wcsncmp (f.dbi.ObjectName.Buffer, L"winpid.", 7) == 0)
+	  status = NtQueryDirectoryObject (dir, dbi_buf, 65536, FALSE, restart,
+					   &context, NULL);
+	  if (!NT_SUCCESS (status))
 	    {
-	      DWORD pid = wcstoul (f.dbi.ObjectName.Buffer + 7, NULL, 10);
-	      add (nelem, false, pid);
+	      debug_printf ("NtQueryDirectoryObject, status %y", status);
+	      break;
+	    }
+	  if (status != STATUS_MORE_ENTRIES)
+	    last_run = true;
+	  restart = FALSE;
+	  for (PDIRECTORY_BASIC_INFORMATION dbi = dbi_buf;
+	       dbi->ObjectName.Length > 0;
+	       dbi++)
+	    {
+	      if (wcsncmp (dbi->ObjectName.Buffer, L"winpid.", 7) == 0)
+		{
+		  DWORD pid = wcstoul (dbi->ObjectName.Buffer + 7, NULL, 10);
+		  add (nelem, false, pid);
+		}
 	    }
 	}
     }
